@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/procmeter3/modules/RCS/acpi.c 1.6 2002/12/07 19:39:22 amb Exp $
+  $Header: /home/amb/procmeter3/modules/RCS/acpi.c 1.16 2007/12/28 15:43:04 amb Exp $
 
-  ProcMeter - A system monitoring program for Linux - Version 3.4.
+  ProcMeter - A system monitoring program for Linux - Version 3.5b.
 
   ACPI source file.
   ******************/ /******************
   Written by Joey Hess.
 
-  This file Copyright 2001 Joey Hess
+  This file Copyright 2001-2007 Joey Hess
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -28,7 +28,7 @@
 #include <dirent.h>
 #include <string.h>
 #ifdef ACPI_APM
-#include <apm.h>
+#include "apm.h"
 #endif
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -36,6 +36,7 @@
 
 #include "acpi.h"
 
+#define PROC_ACPI "/proc/acpi"
 #define ACPI_MAXITEM 8
 
 int acpi_batt_count = 0;
@@ -50,6 +51,48 @@ int acpi_ac_count = 0;
 char acpi_ac_adapter_info[ACPI_MAXITEM][128];
 char acpi_ac_adapter_status[ACPI_MAXITEM][128];
 
+/* These are the strings used in the ACPI shipped with the 2.4 kernels */
+char *acpi_labels_old[] = {
+	"info",
+	"status",
+	"battery",
+	"ac_adapter",
+	"on-line",
+	"Design Capacity:",
+	"Present:",
+	"Remaining Capacity:",
+	"Present Rate:",
+	"State:",
+#if ACPI_THERMAL
+	"thermal",
+#endif
+	"Status:",
+	NULL
+};
+
+/* These are the strings used in ACPI in the 2.5 kernels, circa version
+ * 20020214 */
+char *acpi_labels_20020214[] = {
+	"info",
+	"state",
+	"battery",
+	"ac_adapter",
+	"on-line",
+	"design capacity:",
+	"present:",
+	"remaining capacity:",
+	"present rate:",
+	"charging state:",
+#if ACPI_THERMAL
+	"thermal_zone",
+#endif
+	"state:",
+	"last full capacity:",
+	NULL
+};
+
+char **acpi_labels = NULL;
+
 #if ACPI_THERMAL
 int acpi_thermal_count = 0;
 char acpi_thermal_info[ACPI_MAXITEM][128];
@@ -65,7 +108,7 @@ inline char *get_acpi_file (const char *file) {
 	fd = open(file, O_RDONLY);
 	if (fd == -1) return NULL;
 	end = read(fd, buf, sizeof(buf));
-	buf[end] = '\0';
+	buf[end-1] = '\0';
 	close(fd);
 	return buf;
 }
@@ -87,8 +130,9 @@ inline int scan_acpi_num (const char *buf, const char *key) {
 inline char *scan_acpi_value (const char *buf, const char *key) {
 	char *ptr;
 	static char ret[256];
+	
 	if ((ptr = strstr(buf, key))) {
-		if (sscanf(ptr + strlen(key), "%s", ret) == 1)
+		if (sscanf(ptr + strlen(key), "%255s", ret) == 1)
 			return ret;
 	}
 	return NULL;
@@ -104,54 +148,95 @@ char *get_acpi_value (const char *file, const char *key) {
 	return scan_acpi_value(buf, key);
 }
 
-/* Returns the last full capacity of a battery. */
+/* Returns the maximum capacity of a battery.
+ *
+ * Note that this returns the highest possible capacity for the battery,
+ * even if it can no longer charge that fully. So normally it uses the
+ * design capacity. While the last full capacity of the battery should
+ * never exceed the design capacity, some silly hardware might report
+ * that it does. So if the last full capacity is greater, it will be
+ * returned.
+ */
 int get_acpi_batt_capacity(int battery) {
-	int cap;
-        char *caps=get_acpi_value(acpi_batt_info[battery], "Last Full Capacity:");
-        if (caps == NULL)
-               cap=0; /* battery not present */
-        else
-               cap=atoi(caps);
+	int dcap, lcap;
+	char *s;
+
+	s=get_acpi_value(acpi_batt_info[battery], acpi_labels[label_design_capacity]);
+	if (s == NULL)
+		dcap=0; /* battery not present */
+	else
+		dcap=atoi(s);
+
 	/* This is ACPI's broken way of saying that there is no battery. */
-	if (cap == 655350)
+	if (dcap == 655350)
 		return 0;
-	return cap;
+
+	s=get_acpi_value(acpi_batt_info[battery], acpi_labels[label_last_full_capacity]);
+	if (s != NULL) {
+		lcap=atoi(s);
+		if (lcap > dcap)
+			return lcap;
+	}
+
+	return dcap;
+}
+
+/* Comparison function for qsort. */
+int _acpi_compare_strings (const void *a, const void *b) {
+	const char **pa = (const char **)a;
+	const char **pb = (const char **)b;
+	return strcasecmp((const char *)*pa, (const char *)*pb);
 }
 
 /* Find something (batteries, ac adpaters, etc), and set up a string array
- * to hold the paths to info and status files of the things found. Must be 
- * in /proc/acpi to call this. Returns the number of items found. */
+ * to hold the paths to info and status files of the things found.
+ * Returns the number of items found. */
 int find_items (char *itemname, char infoarray[ACPI_MAXITEM][128],
 		                char statusarray[ACPI_MAXITEM][128]) {
 	DIR *dir;
 	struct dirent *ent;
-	int count = 0;
+	int num_devices=0;
+	int i;
+	char **devices = malloc(ACPI_MAXITEM * sizeof(char *));
 	
-	dir = opendir(itemname);
+	char pathname[128];
+
+	sprintf(pathname,PROC_ACPI "/%s",itemname);
+
+	dir = opendir(pathname);
 	if (dir == NULL)
 		return 0;
-
 	while ((ent = readdir(dir))) {
-		if (!strncmp(".", ent->d_name, 1) || 
-		    !strncmp("..", ent->d_name, 2))
+		if (!strcmp(".", ent->d_name) || 
+		    !strcmp("..", ent->d_name))
 			continue;
 
-		sprintf(infoarray[count], "%s/%s/%s", itemname, ent->d_name, "info");
-		sprintf(statusarray[count], "%s/%s/%s", itemname, ent->d_name, "status");
-		count++;
-		if (count > ACPI_MAXITEM)
+		devices[num_devices]=strdup(ent->d_name);
+		num_devices++;
+		if (num_devices >= ACPI_MAXITEM)
 			break;
 	}
-
 	closedir(dir);
-	return count;
+	
+	/* Sort, since readdir can return in any order. /proc/ does
+	 * sometimes list BATT2 before BATT1. */
+	qsort(devices, num_devices, sizeof(char *), _acpi_compare_strings);
+
+	for (i = 0; i < num_devices; i++) {
+		sprintf(infoarray[i], PROC_ACPI "/%s/%s/%s", itemname, devices[i],
+			acpi_labels[label_info]);
+		sprintf(statusarray[i], PROC_ACPI "/%s/%s/%s", itemname, devices[i],
+			acpi_labels[label_status]);
+		free(devices[i]);
+	}
+
+	return num_devices;
 }
 
 /* Find batteries, return the number, and set acpi_batt_count to it as well. */
 int find_batteries(void) {
 	int i;
-	acpi_batt_count = find_items("battery", acpi_batt_info, acpi_batt_status);
-	/* Read in the last charged capacity of the batteries. */
+	acpi_batt_count = find_items(acpi_labels[label_battery], acpi_batt_info, acpi_batt_status);
 	for (i = 0; i < acpi_batt_count; i++)
 		acpi_batt_capacity[i] = get_acpi_batt_capacity(i);
 	return acpi_batt_count;
@@ -160,7 +245,7 @@ int find_batteries(void) {
 /* Find AC power adapters, return the number found, and set acpi_ac_count to it
  * as well. */
 int find_ac_adapters(void) {
-	acpi_ac_count = find_items("ac_adapter", acpi_ac_adapter_info, acpi_ac_adapter_status);
+	acpi_ac_count = find_items(acpi_labels[label_ac_adapter], acpi_ac_adapter_info, acpi_ac_adapter_status);
 	return acpi_ac_count;
 }
 
@@ -168,7 +253,7 @@ int find_ac_adapters(void) {
 /* Find thermal information sources, return the number found, and set
  * thermal_count to it as well. */
 int find_thermal(void) {
-	acpi_thermal_count = find_items("thermal", acpi_thermal_info, acpi_thermal_status);
+	acpi_thermal_count = find_items(acpi_labels[label_thermal], acpi_thermal_info, acpi_thermal_status);
 	return acpi_thermal_count;
 }
 #endif
@@ -177,7 +262,8 @@ int find_thermal(void) {
 int on_ac_power (void) {
 	int i;
 	for (i = 0; i < acpi_ac_count; i++) {
-		if (strcmp("on-line", get_acpi_value(acpi_ac_adapter_status[i], "Status:")))
+                char *online=get_acpi_value(acpi_ac_adapter_status[i], acpi_labels[label_ac_state]);
+		if (online && strcmp(acpi_labels[label_online], online) == 0)
 			return 1;
 		else
 			return 0;
@@ -189,20 +275,40 @@ int on_ac_power (void) {
  * ac power adapters. */
 int acpi_supported (void) {
 	char *version;
+	DIR *dir;
+	int num;
 
-	if (chdir("/proc/acpi") == -1) {
+	if (!(dir = opendir(PROC_ACPI))) {
 		return 0;
 	}
+	closedir(dir);
 	
-	version = get_acpi_value("info", "ACPI-CA Version:");
+	/* If kernel is 2.6.21 or newer, version is in
+	   /sys/module/acpi/parameters/acpica_version */
+	
+	version = get_acpi_file("/sys/module/acpi/parameters/acpica_version");
+	if (version == NULL) {
+		version = get_acpi_value(PROC_ACPI "/info", "ACPI-CA Version:");
+		if (version == NULL) {
+			/* 2.5 kernel acpi */
+			version = get_acpi_value(PROC_ACPI "/info", "version:");
+		}
+	}
 	if (version == NULL) {
 		return 0;
 	}
-	if (atoi(version) < ACPI_VERSION) {
+	num = atoi(version);
+	if (num < ACPI_VERSION) {
 		fprintf(stderr, "ACPI subsystem %s too is old, consider upgrading to %i.\n",
 				version, ACPI_VERSION);
 		return 0;
 	}
+	else if (num >= 20020214) {
+		acpi_labels = acpi_labels_20020214;
+	}
+	else {
+		acpi_labels = acpi_labels_old;
+	}
 	
 	find_batteries();
 	find_ac_adapters();
@@ -219,51 +325,73 @@ int acpi_supported (void) {
 int acpi_read (int battery, apm_info *info) {
 	char *buf, *state;
 	
+	if (acpi_batt_count == 0) {
+		info->battery_percentage = 0;
+		info->battery_time = 0;
+		info->battery_status = BATTERY_STATUS_ABSENT;
+		acpi_batt_capacity[battery] = 0;
+		/* Where else would the power come from, eh? ;-) */
+		info->ac_line_status = 1;
+		return 0;
+	}
+	
 	/* Internally it's zero indexed. */
 	battery--;
 	
 	buf = get_acpi_file(acpi_batt_status[battery]);
+	if (buf == NULL) {
+		fprintf(stderr, "unable to read %s\n", acpi_batt_status[battery]);
+		perror("read");
+		exit(1);
+	}
 
 	info->ac_line_status = 0;
 	info->battery_flags = 0;
 	info->using_minutes = 1;
 	
-	/* Work out if the battery is present, and what pqercentage of full
+	/* Work out if the battery is present, and what percentage of full
 	 * it is and how much time is left. */
-	if (strcmp(scan_acpi_value(buf, "Present:"), "yes") == 0) {
-		int pcap = scan_acpi_num(buf, "Remaining Capacity:");
-		int rate = scan_acpi_num(buf, "Present Rate:");
+	if (strcmp(scan_acpi_value(buf, acpi_labels[label_present]), "yes") == 0) {
+		int pcap = scan_acpi_num(buf, acpi_labels[label_remaining_capacity]);
+		int rate = scan_acpi_num(buf, acpi_labels[label_present_rate]);
 
 		if (rate) {
 			/* time remaining = (current_capacity / discharge rate) */
 			info->battery_time = (float) pcap / (float) rate * 60;
 		}
 		else {
-			char *rate_s = scan_acpi_value(buf, "Present Rate:");
+			char *rate_s = scan_acpi_value(buf, acpi_labels[label_present_rate]);
 			if (! rate_s) {
 				/* Time remaining unknown. */
 				info->battery_time = 0;
 			}
-			/* This is a hack for my picturebook. If
-			 * the battery is not present, ACPI still
-			 * says it is Present, but sets this to
-			 * unknown. I don't know if this is the
-			 * correct way to do it. */
-			else if (strcmp(rate_s, "unknown") == 0) {
-				goto NOBATT;
+			else {
+				/* a zero or unknown in the file; time 
+				 * unknown so use a negative one to
+				 * indicate this */
+				info->battery_time = -1;
 			}
 		}
 
-		state = scan_acpi_value(buf, "State:");
+		state = scan_acpi_value(buf, acpi_labels[label_charging_state]);
 		if (state) {
 			if (state[0] == 'd') { /* discharging */
 				info->battery_status = BATTERY_STATUS_CHARGING;
+				/* Expensive ac power check used here
+				 * because AC power might be on even if a
+				 * battery is discharging in some cases. */
+				info->ac_line_status = on_ac_power();
 			}
 			else if (state[0] == 'c' && state[1] == 'h') { /* charging */
 				info->battery_status = BATTERY_STATUS_CHARGING;
 				info->ac_line_status = 1;
 				info->battery_flags = info->battery_flags | BATTERY_FLAGS_CHARGING;
-				info->battery_time = -1 * (float) (acpi_batt_capacity[battery] - pcap) / (float) rate * 60;
+				if (rate)
+					info->battery_time = -1 * (float) (acpi_batt_capacity[battery] - pcap) / (float) rate * 60;
+				else
+					info->battery_time = 0;
+				if (abs(info->battery_time) < 0.5)
+					info->battery_time = 0;
 			}
 			else if (state[0] == 'o') { /* ok */
 				/* charged, on ac power */
@@ -272,9 +400,20 @@ int acpi_read (int battery, apm_info *info) {
 			}
 			else if (state[0] == 'c') { /* not charging, so must be critical */
 				info->battery_status = BATTERY_STATUS_CRITICAL;
+				/* Expensive ac power check used here
+				 * because AC power might be on even if a
+				 * battery is critical in some cases. */
+				info->ac_line_status = on_ac_power();
 			}
-			else
+			else if (rate == 0) {
+				/* if rate is null, battery charged, on
+				 * ac power */
+				info->battery_status = BATTERY_STATUS_HIGH;
+				info->ac_line_status = 1;
+			}
+			else {
 				fprintf(stderr, "unknown battery state: %s\n", state);
+			}
 		}
 		else {
 			/* Battery state unknown. */
@@ -289,9 +428,14 @@ int acpi_read (int battery, apm_info *info) {
 			 * buf below this point! */
 			acpi_batt_capacity[battery] = get_acpi_batt_capacity(battery);
 		}
+		else if (pcap > acpi_batt_capacity[battery]) {
+			/* Battery is somehow charged to greater than max
+			 * capacity. Rescan for a new max capacity. */
+			find_batteries();
+		}
 		
-		if (pcap) {
-			/* percentage = (current_capacity / last_full_capacity) * 100 */
+		if (pcap && acpi_batt_capacity[battery]) {
+			/* percentage = (current_capacity / max capacity) * 100 */
 			info->battery_percentage = (float) pcap / (float) acpi_batt_capacity[battery] * 100;
 		}
 		else {
@@ -300,12 +444,11 @@ int acpi_read (int battery, apm_info *info) {
 
 	}
 	else {
-NOBATT:
 		info->battery_percentage = 0;
 		info->battery_time = 0;
 		info->battery_status = BATTERY_STATUS_ABSENT;
 		acpi_batt_capacity[battery] = 0;
-		if (acpi_batt_count == 1) {
+		if (acpi_batt_count == 0) {
 			/* Where else would the power come from, eh? ;-) */
 			info->ac_line_status = 1;
 		}
@@ -375,13 +518,13 @@ ProcMeterOutput _batt_outputs[N_BATT_OUTPUTS]=
  /*+ Rate of charge/discharge. +*/
  {
   /* char  name[];          */ "Batt%i_Rate",
-  /* char *description;     */ "The present rate of battery %i discharge or charge, in mW.",
+  /* char *description;     */ "The present rate of battery %i discharge or charge, in mA.",
   /* char  type;            */ PROCMETER_GRAPH|PROCMETER_TEXT|PROCMETER_BAR,
   /* short interval;        */ 1,
-  /* char  text_value[];    */ "0 mW",
+  /* char  text_value[];    */ "0 mA",
   /* long  graph_value;     */ 0,
   /* short graph_scale;     */ 1000,
-  /* char  graph_Units[];   */ "(%d mW)"
+  /* char  graph_Units[];   */ "(%d mA)"
  },
  /*+ The battery status output +*/
  {
@@ -572,16 +715,16 @@ int Update(time_t now, ProcMeterOutput *output)
 		buf = get_acpi_file(acpi_batt_status[index]);
 		if (! buf) return(-1);
 		
-		if (strcmp(scan_acpi_value(buf, "Present:"), "yes") == 0) {
-			pcap = scan_acpi_num(buf, "Remaining Capacity:");
-			rate = scan_acpi_num(buf, "Present Rate:");
+		if (strcmp(scan_acpi_value(buf, acpi_labels[label_present]), "yes") == 0) {
+			pcap = scan_acpi_num(buf, acpi_labels[label_remaining_capacity]);
+			rate = scan_acpi_num(buf, acpi_labels[label_present_rate]);
 			
 			if (rate) {
 				/* time remaining till empty = current_capacity / discharge rate) */
 				timeleft = (float) pcap / (float) rate * 60;
 			}
 			else {
-				char *rate_s = scan_acpi_value(buf, "Present Rate:");
+				char *rate_s = scan_acpi_value(buf, acpi_labels[label_present_rate]);
 				/* If the battery is not present, ACPI may
 				 * still say it is but sets rate to unknown
 				 * (on my picturebook, anyway). I don't
@@ -596,7 +739,7 @@ int Update(time_t now, ProcMeterOutput *output)
 			timefull = (float)(acpi_batt_capacity[index] - pcap) / (float) rate * 60;
 
 			/* Status. */
-			status = scan_acpi_value(buf, "State:");
+			status = scan_acpi_value(buf, acpi_labels[label_charging_state]);
 			sprintf(batt_outputs[index + 2].text_value, "%s", status);
 			
 			if (strcmp(status, "charging") == 0) {
@@ -653,7 +796,7 @@ NOBATT:
 	
 		/* Charge/discharge rate. */
 		batt_outputs[index + 1].graph_value = PROCMETER_GRAPH_FLOATING((float)rate/batt_outputs[index + 1].graph_scale);
-		sprintf(batt_outputs[index + 1].text_value,"%i mW", rate);
+		sprintf(batt_outputs[index + 1].text_value,"%i mA", rate);
 	}
 
 	return(0);
