@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/procmeter3/modules/RCS/stat-disk.c 1.11 2004/04/03 16:08:24 amb Exp $
+  $Header: /home/amb/procmeter3/modules/RCS/stat-disk.c 1.15 2008/12/07 17:35:47 amb Exp $
 
-  ProcMeter - A system monitoring program for Linux - Version 3.4b.
+  ProcMeter - A system monitoring program for Linux - Version 3.5c.
 
   Disk statistics source file.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1998,99,2000,02,04 Andrew M. Bishop
+  This file Copyright 1998-2008 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -18,6 +18,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 
 #include <unistd.h>
 #include <dirent.h>
@@ -32,9 +33,6 @@
 #define DISK_WRITE  2
 #define N_OUTPUTS   3
 
-/*+ The length of the buffer for reading in lines. +*/
-#define BUFFLEN 2048
-
 /* The interface information.  */
 
 /*+ The total disk statistics +*/
@@ -161,22 +159,29 @@ ProcMeterOutput **outputs=NULL;
 ProcMeterModule module=
 {
  /* char name[];            */ "Stat-Disk",
- /* char *description;      */ "Disk usage statistics. [From /proc/stat]",
+ /* char *description;      */ "Disk usage statistics. [From /proc/diskstats or /proc/stat]",
 };
 
 
-static int add_disk(char *devname);
+/* The line buffer */
+static char *line=NULL;
+static size_t length=0;
 
-static unsigned long *current,*previous,*values[2];
+/* The current and previous information about the disk */
+static unsigned long long *current,*previous,*values[2];
 
+/* The information about the disks */
+static int ndisks=4;
 static unsigned *majors=NULL,*minors=NULL,*indexes=NULL;
 
-static int ndisks=4;
-
+/* The estimated kernel version based on the file format */
 static int kernel_version_130=0;
 static int kernel_version_240=0;
 static int kernel_version_260=0;
 
+/* A function to add a disk */
+static int add_disk(char *devname);
+
 
 /*++++++++++++++++++++++++++++++++++++++
   Load the module.
@@ -201,7 +206,6 @@ ProcMeterModule *Load(void)
 ProcMeterOutput **Initialise(char *options)
 {
  FILE *f;
- char line[BUFFLEN+1],*l;
 
  disk_outputs=(ProcMeterOutput*)malloc((N_OUTPUTS*ndisks)*sizeof(ProcMeterOutput));
 
@@ -221,36 +225,36 @@ ProcMeterOutput **Initialise(char *options)
        fprintf(stderr,"ProcMeter(%s): Could not open '/proc/stat'.\n",__FILE__);
     else
       {
-       l=fgets(line,256,f); /* cpu */
-       if(!l)
+       /* cpu */
+       if(!fgets_realloc(&line,&length,f))
           fprintf(stderr,"ProcMeter(%s): Could not read '/proc/stat'.\n",__FILE__);
        else
          {
-          while(l && !(line[0]=='d' && line[1]=='i' && line[2]=='s' && line[3]=='k'))
-             l=fgets(line,BUFFLEN,f); /* cpu or disk or page or swap or intr or disk_io */
+          while(fgets_realloc(&line,&length,f)) /* cpu or disk or page or swap or intr or disk_io */
+             if(line[0]=='d' && line[1]=='i' && line[2]=='s' && line[3]=='k')
+                break;
 
-          if(!l)
+          if(!line[0])
              fprintf(stderr,"ProcMeter(%s): Unexpected 'disk' line in '/proc/stat'.\n"
                      "    expected: 'disk ...' or 'disk_io ...'\n"
                      "    found:    EOF\n",__FILE__);
           else if(!strncmp(line,"disk ",5)) /* kernel version < ~2.4.0-test4 */
             {
-             unsigned long d1,d2,d3,d4;
+             unsigned long long d1,d2,d3,d4;
 
-             if(sscanf(line,"disk %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+             if(sscanf(line,"disk %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
                {
                 int i,j;
                 int read_avail=0,write_avail=0;
                 int n=0;
 
-                l=fgets(line,BUFFLEN,f); /* disk_* or page */
-                while(l && line[0]=='d') /* kernel version > ~1.3.0 */
+                /* disk_* or page */
+                while(fgets_realloc(&line,&length,f) && line[0]=='d') /* kernel version > ~1.3.0 */
                   {
-                   if(sscanf(line,"disk_rblk %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+                   if(sscanf(line,"disk_rblk %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
                       read_avail=1;
-                   if(sscanf(line,"disk_wblk %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+                   if(sscanf(line,"disk_wblk %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
                       write_avail=1;
-                   l=fgets(line,BUFFLEN,f); /* disk_* or page */
                   }
 
                 if(read_avail && write_avail)
@@ -284,8 +288,8 @@ ProcMeterOutput **Initialise(char *options)
           else if(!strncmp(line,"disk_io: ",9)) /* kernel version > ~2.4.0-test4 */
             {
              int maj,min,idx,num=8,nm,nr;
-             unsigned long d1,d2,d3,d4,d5;
-             DIR *devdir,*devdiscsdir;
+             unsigned long long d1,d2,d3,d4,d5;
+             DIR *devdir,*devdiscsdir,*devmapperdir;
              char devname[PATH_MAX];
              struct dirent *ent;
              struct stat buf;
@@ -296,9 +300,10 @@ ProcMeterOutput **Initialise(char *options)
 
              devdir=opendir("/dev");
              devdiscsdir=opendir("/dev/discs");
+             devmapperdir=opendir("/dev/mapper");
 
-             while((nr=sscanf(line+num," (%d,%d):(%lu,%lu,%lu,%lu,%lu)%n",&maj,&idx,&d1,&d2,&d3,&d4,&d5,&nm))==7 ||
-                   (nr=sscanf(line+num," (%d,%d):(%lu,%lu,%lu,%lu)%n",&maj,&idx,&d1,&d2,&d3,&d4,&nm))==6)
+             while((nr=sscanf(line+num," (%d,%d):(%llu,%llu,%llu,%llu,%llu)%n",&maj,&idx,&d1,&d2,&d3,&d4,&d5,&nm))==7 ||
+                   (nr=sscanf(line+num," (%d,%d):(%llu,%llu,%llu,%llu)%n",&maj,&idx,&d1,&d2,&d3,&d4,&nm))==6)
                {
                 int done=0;
 
@@ -327,7 +332,18 @@ ProcMeterOutput **Initialise(char *options)
                    min=0;
                   }
 
-                if(devdiscsdir)
+                if(devmapperdir)
+                  {
+                   rewinddir(devmapperdir);
+                   while((ent=readdir(devmapperdir)))
+                     {
+                      sprintf(devname,"/dev/mapper/%s",ent->d_name);
+                      if(!lstat(devname,&buf) && S_ISBLK(buf.st_mode) && !S_ISLNK(buf.st_mode) &&
+                         major(buf.st_rdev)==maj && minor(buf.st_rdev)==min)
+                         done=add_disk(devname);
+                     }
+                  }
+                if(!done && devdiscsdir)
                   {
                    rewinddir(devdiscsdir);
                    while((ent=readdir(devdiscsdir)))
@@ -351,7 +367,7 @@ ProcMeterOutput **Initialise(char *options)
                   }
                 if(!done)
                   {
-                   fprintf(stderr,"ProcMeter(%s): Cannot find disk in /dev or /dev/discs for device %d:%d in '/proc/stat'.\n",__FILE__,maj,min);
+                   fprintf(stderr,"ProcMeter(%s): Cannot find disk in /dev, /dev/mapper, or /dev/discs for device %d:%d in '/proc/stat'.\n",__FILE__,maj,min);
                    continue;
                   }
                }
@@ -372,14 +388,13 @@ ProcMeterOutput **Initialise(char *options)
    }
  else /* Found /proc/diskstats in kernel 2.6.x */
    {
-    l=fgets(line,256,f); /* cpu */
-    if(!l)
+    if(!fgets_realloc(&line,&length,f))
        fprintf(stderr,"ProcMeter(%s): Could not read '/proc/diskstats'.\n",__FILE__);
     else
       {
        int maj,min,nr;
-       unsigned long d1,d2,d3,d4,d5;
-       DIR *devdir,*devdiscsdir;
+       unsigned long long d1,d2,d3,d4,d5;
+       DIR *devdir,*devdiscsdir,*devmapperdir;
        char devname[PATH_MAX];
        struct dirent *ent;
        struct stat buf;
@@ -390,26 +405,38 @@ ProcMeterOutput **Initialise(char *options)
 
        devdir=opendir("/dev");
        devdiscsdir=opendir("/dev/discs");
+       devmapperdir=opendir("/dev/mapper");
 
        do
          {
           int done=0;
 
-          nr=sscanf(line,"%d %d %*s %lu %lu %lu %lu %lu",&maj,&min,&d1,&d2,&d3,&d4,&d5);
+          nr=sscanf(line,"%d %d %*s %llu %llu %llu %llu %llu",&maj,&min,&d1,&d2,&d3,&d4,&d5);
 
           if(nr<6)
              break;
 
-          if(devdiscsdir)
+          if(devmapperdir)
             {
-             rewinddir(devdiscsdir);
-             while((ent=readdir(devdiscsdir)))
-               {
-                sprintf(devname,"/dev/discs/%s/disc",ent->d_name);
-                if(!lstat(devname,&buf) && S_ISBLK(buf.st_mode) && !S_ISLNK(buf.st_mode) &&
-                   major(buf.st_rdev)==maj && minor(buf.st_rdev)==min)
-                   done=add_disk(devname);
-               }
+              rewinddir(devmapperdir);
+              while((ent=readdir(devmapperdir)))
+                {
+                  sprintf(devname,"/dev/mapper/%s",ent->d_name);
+                  if(!lstat(devname,&buf) && S_ISBLK(buf.st_mode) && !S_ISLNK(buf.st_mode) &&
+                     major(buf.st_rdev)==maj && minor(buf.st_rdev)==min)
+                    done=add_disk(devname);
+                }
+            }
+          if(!done && devdiscsdir)
+            {
+              rewinddir(devdiscsdir);
+              while((ent=readdir(devdiscsdir)))
+                {
+                  sprintf(devname,"/dev/discs/%s/disc",ent->d_name);
+                  if(!lstat(devname,&buf) && S_ISBLK(buf.st_mode) && !S_ISLNK(buf.st_mode) &&
+                     major(buf.st_rdev)==maj && minor(buf.st_rdev)==min)
+                    done=add_disk(devname);
+                }
             }
           if(!done && devdir)
             {
@@ -424,11 +451,11 @@ ProcMeterOutput **Initialise(char *options)
             }
           if(!done)
             {
-             fprintf(stderr,"ProcMeter(%s): Cannot find disk in /dev or /dev/discs for device %d:%d in '/proc/diskstats'.\n",__FILE__,maj,min);
+             fprintf(stderr,"ProcMeter(%s): Cannot find disk in /dev, /dev/mapper, or /dev/discs for device %d:%d in '/proc/diskstats'.\n",__FILE__,maj,min);
              continue;
             }
          }
-       while(fgets(line,256,f));
+       while(fgets_realloc(&line,&length,f));
 
        if(devdir)
           closedir(devdir);
@@ -484,8 +511,8 @@ ProcMeterOutput **Initialise(char *options)
     outputs[n]=NULL;
    }
 
- values[0]=(unsigned long*)calloc((N_OUTPUTS*(ndisks+1)),sizeof(unsigned long));
- values[1]=(unsigned long*)calloc((N_OUTPUTS*(ndisks+1)),sizeof(unsigned long));
+ values[0]=(unsigned long long*)calloc((N_OUTPUTS*(ndisks+1)),sizeof(unsigned long long));
+ values[1]=(unsigned long long*)calloc((N_OUTPUTS*(ndisks+1)),sizeof(unsigned long long));
 
  current=values[0];
  previous=values[1];
@@ -593,8 +620,7 @@ int Update(time_t now,ProcMeterOutput *output)
  if(now!=last)
    {
     FILE *f;
-    char line[BUFFLEN+1],*l;
-    long *temp;
+    unsigned long long *temp;
 
     temp=current;
     current=previous;
@@ -609,13 +635,13 @@ int Update(time_t now,ProcMeterOutput *output)
        current[DISK_READ]=0;
        current[DISK_WRITE]=0;
 
-       while((l=fgets(line,BUFFLEN,f)))
+       while(fgets_realloc(&line,&length,f))
          {
           int maj,min,nr;
-          unsigned long d1,d2,d3,d4,d5,dr,dw;
+          unsigned long long d1,d2,d3,d4,d5,dr,dw;
           int j;
 
-          nr=sscanf(line,"%d %d %*s %lu %lu %lu %lu %lu",&maj,&min,&d1,&d2,&d3,&d4,&d5);
+          nr=sscanf(line,"%d %d %*s %llu %llu %llu %llu %llu",&maj,&min,&d1,&d2,&d3,&d4,&d5);
 
           if(nr==7)
              dr=d1,dw=d5;
@@ -633,7 +659,36 @@ int Update(time_t now,ProcMeterOutput *output)
                 current[N_OUTPUTS*(j+1)+DISK]=dr+dw;
                }
 
-          if(nr==7) /* Count the disks, not the partitions */
+          /* Add up the values to get the total, but ignore any 'whole disk' devices */
+
+          if(((maj==IDE0_MAJOR || maj==IDE1_MAJOR || maj==IDE2_MAJOR || maj==IDE3_MAJOR ||
+               maj==IDE4_MAJOR || maj==IDE5_MAJOR || maj==IDE6_MAJOR || maj==IDE7_MAJOR ||
+               maj==IDE8_MAJOR || maj==IDE9_MAJOR) && (min%64)==0) ||
+             ((maj==SCSI_DISK0_MAJOR || (maj>=SCSI_DISK1_MAJOR && maj<=SCSI_DISK7_MAJOR) ||
+               (maj>=SCSI_DISK8_MAJOR && maj<=SCSI_DISK15_MAJOR)) && (min%16)==0) ||
+             (maj==XT_DISK_MAJOR    && (min%64)==0) ||
+             (maj==14               && (min%64)==0) ||
+             (maj==MFM_ACORN_MAJOR  && (min%64)==0) ||
+             (maj==ACSI_MAJOR       && (min%16)==0) ||
+             (maj==PS2ESDI_MAJOR    && (min%64)==0) ||
+             (maj==44               && (min%16)==0) ||
+             (maj==45               && (min%16)==0) ||
+             (maj>=48 && maj<=55    && (min%8)==0)  ||
+             (maj>=COMPAQ_SMART2_MAJOR && maj<=COMPAQ_SMART2_MAJOR7 && (min%16)==0) ||
+             (maj>=I2O_MAJOR && maj<=(I2O_MAJOR+7) && (min%16)==0) ||
+             (maj==92               && (min%16)==0) ||
+             (maj==UBD_MAJOR        && (min%16)==0) ||
+             (maj==101              && (min%16)==0) ||
+             (maj==102              && (min%16)==0) ||
+             (maj>=COMPAQ_CISS_MAJOR && maj<=COMPAQ_CISS_MAJOR7 && (min%16)==0) ||
+             (maj==VIODASD_MAJOR    && (min%8)==0)  ||
+             (maj==ATARAID_MAJOR    && (min%16)==0) ||
+             (maj>=136 && maj<=143  && (min%8)==0)  ||
+             (maj==153              && (min%16)==0) ||
+             (maj>=160 && maj<=161  && (min%32)==0) ||
+             (maj==XENVBD_MAJOR     && (min%16)==0))
+             ;
+          else
             {
              current[DISK_READ] +=dr;
              current[DISK_WRITE]+=dw;
@@ -650,7 +705,7 @@ int Update(time_t now,ProcMeterOutput *output)
        if(!f)
           return(-1);
 
-       while((l=fgets(line,BUFFLEN,f))) /* cpu or disk or page or swap or intr or disk_io */
+       while(fgets_realloc(&line,&length,f)) /* cpu or disk or page or swap or intr or disk_io */
           if(line[0]=='d' && line[1]=='i' && line[2]=='s' && line[3]=='k')
              break;
 
@@ -658,25 +713,23 @@ int Update(time_t now,ProcMeterOutput *output)
          {
           if(!kernel_version_130)
             {
-             sscanf(line,"disk %lu %lu %lu %lu",&current[N_OUTPUTS*1+DISK],&current[N_OUTPUTS*2+DISK],
-                                                &current[N_OUTPUTS*3+DISK],&current[N_OUTPUTS*4+DISK]);
+             sscanf(line,"disk %llu %llu %llu %llu",&current[N_OUTPUTS*1+DISK],&current[N_OUTPUTS*2+DISK],
+                                                    &current[N_OUTPUTS*3+DISK],&current[N_OUTPUTS*4+DISK]);
              current[DISK]=current[N_OUTPUTS*1+DISK]+current[N_OUTPUTS*2+DISK]+
                            current[N_OUTPUTS*3+DISK]+current[N_OUTPUTS*4+DISK];
             }
 
-          l=fgets(line,BUFFLEN,f); /* disk_* or page */
-
-          while(l && line[0]=='d') /* kernel version > ~1.3.0 */
+          /* disk_* or page */
+          while(fgets_realloc(&line,&length,f) && line[0]=='d') /* kernel version > ~1.3.0 */
             {
-             if(sscanf(line,"disk_rblk %lu %lu %lu %lu",&current[N_OUTPUTS*1+DISK_READ],&current[N_OUTPUTS*2+DISK_READ],
-                                                        &current[N_OUTPUTS*3+DISK_READ],&current[N_OUTPUTS*4+DISK_READ])==4)
+             if(sscanf(line,"disk_rblk %llu %llu %llu %llu",&current[N_OUTPUTS*1+DISK_READ],&current[N_OUTPUTS*2+DISK_READ],
+                                                            &current[N_OUTPUTS*3+DISK_READ],&current[N_OUTPUTS*4+DISK_READ])==4)
                 current[DISK_READ]=current[N_OUTPUTS*1+DISK_READ]+current[N_OUTPUTS*2+DISK_READ]+
                                    current[N_OUTPUTS*3+DISK_READ]+current[N_OUTPUTS*4+DISK_READ];
-             if(sscanf(line,"disk_wblk %lu %lu %lu %lu",&current[N_OUTPUTS*1+DISK_WRITE],&current[N_OUTPUTS*2+DISK_WRITE],
-                                                        &current[N_OUTPUTS*3+DISK_WRITE],&current[N_OUTPUTS*4+DISK_WRITE])==4)
+             if(sscanf(line,"disk_wblk %llu %llu %llu %llu",&current[N_OUTPUTS*1+DISK_WRITE],&current[N_OUTPUTS*2+DISK_WRITE],
+                                                            &current[N_OUTPUTS*3+DISK_WRITE],&current[N_OUTPUTS*4+DISK_WRITE])==4)
                 current[DISK_WRITE]=current[N_OUTPUTS*1+DISK_WRITE]+current[N_OUTPUTS*2+DISK_WRITE]+
                                     current[N_OUTPUTS*3+DISK_WRITE]+current[N_OUTPUTS*4+DISK_WRITE];
-             fgets(line,BUFFLEN,f); /* disk_* or page */
             }
 
           if(kernel_version_130)
@@ -698,12 +751,12 @@ int Update(time_t now,ProcMeterOutput *output)
           while(1)
             {
              unsigned maj,idx;
-             unsigned long d1,d3;
+             unsigned long long d1,d3;
 
              if(kernel_version_240==6)
-                nr=sscanf(line+num," (%d,%d):(%*u,%lu,%*u,%lu)%n",&maj,&idx,&d1,&d3,&nm);
+                nr=sscanf(line+num," (%d,%d):(%*u,%llu,%*u,%llu)%n",&maj,&idx,&d1,&d3,&nm);
              else if(kernel_version_240==7)
-                nr=sscanf(line+num," (%d,%d):(%*u,%lu,%*u,%lu,%*u)%n",&maj,&idx,&d1,&d3,&nm);
+                nr=sscanf(line+num," (%d,%d):(%*u,%llu,%*u,%llu,%*u)%n",&maj,&idx,&d1,&d3,&nm);
 
              if(nr!=4)
                 break;
@@ -770,4 +823,7 @@ void Unload(void)
 
  free(values[0]);
  free(values[1]);
+
+ if(line)
+    free(line);
 }
