@@ -1,13 +1,13 @@
 /***************************************
-  $Header: /home/amb/procmeter3/modules/RCS/stat.c 1.14 2004/04/03 16:06:41 amb Exp $
+  $Header: /home/amb/procmeter3/modules/RCS/stat.c 1.18 2008/05/05 18:45:36 amb Exp $
 
-  ProcMeter - A system monitoring program for Linux - Version 3.4b.
+  ProcMeter - A system monitoring program for Linux - Version 3.5b.
 
   Low level system statistics source file.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1998,99,2000,02,04 Andrew M. Bishop
+  This file Copyright 1998-2008 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -38,9 +38,6 @@
 #define INTR       15
 #define N_OUTPUTS  16
 
-/*+ The length of the buffer for reading in lines. +*/
-#define BUFFLEN 2048
-
 /* The interface information.  */
 
 /*+ The statistics +*/
@@ -235,11 +232,20 @@ ProcMeterModule module=
 };
 
 
-static int available[N_OUTPUTS];
-static unsigned long *current,*previous,values[2][N_OUTPUTS];
+/* The line buffer */
+static char *line=NULL;
+static size_t length=0;
 
+/* Information about the availability and the current and previous values */
+static int available[N_OUTPUTS];
+static unsigned long long *current,*previous,values[2][N_OUTPUTS];
+
+/*+ A flag to indicate that this is kernel v2.4 and has many disk IO counters. +*/
 static int kernel_version_240=0;
 
+/*+ A flag to indicate that this is kernel v2.6 and has 8 CPU counters. +*/
+static int kernel_26=0;
+
 
 /*++++++++++++++++++++++++++++++++++++++
   Load the module.
@@ -264,7 +270,6 @@ ProcMeterModule *Load(void)
 ProcMeterOutput **Initialise(char *options)
 {
  FILE *f;
- char line[BUFFLEN+1],*l;
  int n;
 
  outputs[0]=NULL;
@@ -282,89 +287,91 @@ ProcMeterOutput **Initialise(char *options)
     fprintf(stderr,"ProcMeter(%s): Could not open '/proc/stat'.\n",__FILE__);
  else
    {
-    l=fgets(line,BUFFLEN,f);
-    if(!l) /* cpu */
+    if(!fgets_realloc(&line,&length,f)) /* cpu */
        fprintf(stderr,"ProcMeter(%s): Could not read '/proc/stat'.\n",__FILE__);
     else
       {
        int i;
-       unsigned long d1,d2,d3,d4;
+       unsigned long long d1,d2,d3,d4,d5,d6,d7,d8;
 
-       if(sscanf(line,"cpu %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+       if(sscanf(line,"cpu %llu %llu %llu %llu %llu %llu %llu %llu",&d1,&d2,&d3,&d4,&d5,&d6,&d7,&d8)==8)
+          kernel_26=1;
+ 
+       if(kernel_26 || sscanf(line,"cpu %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
           available[CPU]=available[CPU_USER]=available[CPU_NICE]=available[CPU_SYS]=available[CPU_IDLE]=1;
        else
           fprintf(stderr,"ProcMeter(%s): Unexpected 'cpu' line in '/proc/stat'.\n"
-                  "    expected: 'cpu %%lu %%lu %%lu %%lu'\n"
+                  "    expected: 'cpu %%llu %%llu %%llu %%llu'\n"
+                  "          or: 'cpu %%llu %%llu %%llu %%llu %%llu %%llu %%llu %%llu'\n"
                   "    found:    %s",__FILE__,line);
 
-       while(l && (line[0]=='c' && line[1]=='p' && line[2]=='u')) /* kernel version > ~2.1.84 */
-          l=fgets(line,BUFFLEN,f); /* cpu or disk or page or intr */
+       /* cpu or disk or page or intr */
+       while(fgets_realloc(&line,&length,f) && (line[0]=='c' && line[1]=='p' && line[2]=='u')) /* kernel version > ~2.1.84 */
+          ;
 
        if(!strncmp(line,"disk ",5)) /* kernel version < ~2.4.0-test4 */
          {
-          unsigned long d1,d2,d3,d4;
+          unsigned long long d1,d2,d3,d4;
 
-          if(sscanf(line,"disk %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+          if(sscanf(line,"disk %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
              available[DISK]=1;
           else
              fprintf(stderr,"ProcMeter(%s): Unexpected 'disk' line in '/proc/stat'.\n"
-                            "    expected: 'disk %%u %%u %%u %%u'\n"
+                            "    expected: 'disk %%llu %%llu %%llu %%llu'\n"
                             "    found:    %s",__FILE__,line);
 
-          l=fgets(line,BUFFLEN,f); /* disk_* or page */
-
-          while(l && line[0]=='d') /* kernel version > ~1.3.0 */
+          /* disk_* or page */
+          while(fgets_realloc(&line,&length,f) && line[0]=='d') /* kernel version > ~1.3.0 */
             {
-             if(sscanf(line,"disk_rblk %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+             if(sscanf(line,"disk_rblk %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
                 available[DISK_READ]=1;
-             if(sscanf(line,"disk_wblk %lu %lu %lu %lu",&d1,&d2,&d3,&d4)==4)
+             if(sscanf(line,"disk_wblk %llu %llu %llu %llu",&d1,&d2,&d3,&d4)==4)
                 available[DISK_WRITE]=1;
-             l=fgets(line,BUFFLEN,f); /* disk_* or page */
             }
          }
 
        if(!strncmp(line,"page",4)) /* kernel version < ~2.5.30 */
          {
-          if(sscanf(line,"page %lu %lu",&d1,&d2)==2)
+          if(sscanf(line,"page %llu %llu",&d1,&d2)==2)
             {
              available[PAGE]=available[PAGE_IN]=available[PAGE_OUT]=1;
-             l=fgets(line,BUFFLEN,f); /* swap or intr */
+             fgets_realloc(&line,&length,f); /* swap or intr */
             }
           else
              fprintf(stderr,"ProcMeter(%s): Unexpected 'page' line in '/proc/stat'.\n"
-                            "    expected: 'page %%lu %%lu'\n"
+                            "    expected: 'page %%llu %%llu'\n"
                             "    found:    %s",__FILE__,line);
 
-          if(sscanf(line,"swap %lu %lu",&d1,&d2)==2)
+          if(sscanf(line,"swap %llu %llu",&d1,&d2)==2)
             {
              available[SWAP]=available[SWAP_IN]=available[SWAP_OUT]=1;
-             l=fgets(line,BUFFLEN,f); /* intr */
+             fgets_realloc(&line,&length,f); /* intr */
             }
           else
              fprintf(stderr,"ProcMeter(%s): Unexpected 'swap' line in '/proc/stat'.\n"
-                            "    expected: 'swap %%lu %%lu'\n"
+                            "    expected: 'swap %%llu %%llu'\n"
                             "    found:    %s",__FILE__,line);
          }
 
-       if(sscanf(line,"intr %lu",&d1)==1)
+       if(sscanf(line,"intr %llu",&d1)==1)
          {
           available[INTR]=1;
-          l=fgets(line,BUFFLEN,f); /* disk_io or ctxt */
+          fgets_realloc(&line,&length,f); /* disk_io or ctxt */
          }
        else
           fprintf(stderr,"ProcMeter(%s): Unexpected 'intr' line in '/proc/stat'.\n"
-                         "    expected: 'intr %%lu ...'\n"
+                         "    expected: 'intr %%llu ...'\n"
                          "    found:    %s",__FILE__,line);
 
        if(!strncmp(line,"disk_io: ",9)) /* kernel version > ~2.4.0-test4 */
          {
           int maj,idx,num=8,nm,nr;
-          unsigned long d0,d1,d2,d3,d4;
+          unsigned long long d0,d1,d2,d3,d4;
 
           kernel_version_240=1;
 
-          while((nr=sscanf(line+num," (%d,%d):(%lu,%lu,%lu,%lu,%lu)%n",&maj,&idx,&d0,&d1,&d2,&d3,&d4,&nm))==7 ||
-                (nr=sscanf(line+num," (%d,%d):(%lu,%lu,%lu,%lu)%n",&maj,&idx,&d0,&d1,&d2,&d3,&nm))==6)
+          while((nr=sscanf(line+num," (%d,%d):(%llu,%llu,%llu,%llu,%llu)%n",&maj,&idx,&d0,&d1,&d2,&d3,&d4,&nm))==7 ||
+                (nr=sscanf(line+num," (%d,%d):(%llu,%llu,%llu,%llu)%n",&maj,&idx,&d0,&d1,&d2,&d3,&nm))==6)
             {
              num+=nm;
 
@@ -375,14 +382,14 @@ ProcMeterOutput **Initialise(char *options)
              available[DISK_WRITE]=1;
             }
 
-          l=fgets(line,BUFFLEN,f); /* ctxt */
+          fgets_realloc(&line,&length,f); /* ctxt */
          }
 
-       if(sscanf(line,"ctxt %lu",&d1)==1)
+       if(sscanf(line,"ctxt %llu",&d1)==1)
           available[CONTEXT]=1;
        else
           fprintf(stderr,"ProcMeter(%s): Unexpected 'ctxt' line in '/proc/stat'.\n"
-                         "    expected: 'ctxt %%lu'\n"
+                         "    expected: 'ctxt %%llu'\n"
                          "    found:    %s",__FILE__,line);
 
        for(i=0;i<N_OUTPUTS;i++)
@@ -420,8 +427,8 @@ int Update(time_t now,ProcMeterOutput *output)
  if(now!=last)
    {
     FILE *f;
-    char line[BUFFLEN+1],*l;
-    long *temp;
+    unsigned long long *temp;
+    unsigned long long cpu_iowait,cpu_irq,cpu_softirq,cpu_steal;
 
     temp=current;
     current=previous;
@@ -431,56 +438,56 @@ int Update(time_t now,ProcMeterOutput *output)
     if(!f)
        return(-1);
 
-    l=fgets(line,BUFFLEN,f); /* cpu */
+    fgets_realloc(&line,&length,f); /* cpu */
     if(available[CPU])
-       sscanf(line,"cpu %lu %lu %lu %lu",&current[CPU_USER],&current[CPU_NICE],&current[CPU_SYS],&current[CPU_IDLE]);
+       sscanf(line,"cpu %llu %llu %llu %llu %llu %llu %llu %llu",&current[CPU_USER],&current[CPU_NICE],&current[CPU_SYS],&current[CPU_IDLE],
+                                                                 &cpu_iowait,&cpu_irq,&cpu_softirq,&cpu_steal);
 
-    while(l && line[0]=='c') /* kernel version > ~2.1.84 */
-       l=fgets(line,BUFFLEN,f); /* cpu or disk or page or intr */
+    /* cpu or disk or page or intr */
+    while(fgets_realloc(&line,&length,f) && line[0]=='c') /* kernel version > ~2.1.84 */
+       ;
 
     if(available[DISK] && !kernel_version_240) /* kernel version < ~2.4.0-test4 */
       {
-       unsigned long d0,d1,d2,d3;
+       unsigned long long d0,d1,d2,d3;
 
        if(available[DISK])
-         {sscanf(line,"disk %lu %lu %lu %lu",&d0,&d1,&d2,&d3);current[DISK]=d0+d1+d2+d3;}
+         {sscanf(line,"disk %llu %llu %llu %llu",&d0,&d1,&d2,&d3);current[DISK]=d0+d1+d2+d3;}
 
-       l=fgets(line,BUFFLEN,f); /* disk_* or page */
-
-       while(line[0]=='d') /* kernel version > ~1.3.x */
+       /* disk_* or page */
+       while(fgets_realloc(&line,&length,f) && line[0]=='d') /* kernel version > ~1.3.x */
          {
           if(available[DISK_READ])
-             if(sscanf(line,"disk_rblk %lu %lu %lu %lu",&d0,&d1,&d2,&d3)==4)
+             if(sscanf(line,"disk_rblk %llu %llu %llu %llu",&d0,&d1,&d2,&d3)==4)
                {current[DISK_READ]=d0+d1+d2+d3;}
           if(available[DISK_WRITE])
-             if(sscanf(line,"disk_wblk %lu %lu %lu %lu",&d0,&d1,&d2,&d3)==4)
+             if(sscanf(line,"disk_wblk %llu %llu %llu %llu",&d0,&d1,&d2,&d3)==4)
                {current[DISK_WRITE]=d0+d1+d2+d3;}
-          l=fgets(line,BUFFLEN,f); /* disk_* or page */
          }
       }
 
     if(available[PAGE])
       {
-       sscanf(line,"page %lu %lu",&current[PAGE_IN],&current[PAGE_OUT]);
-       l=fgets(line,BUFFLEN,f); /* swap */
+       sscanf(line,"page %llu %llu",&current[PAGE_IN],&current[PAGE_OUT]);
+       fgets_realloc(&line,&length,f); /* swap */
       }
 
     if(available[SWAP])
       {
-       sscanf(line,"swap %lu %lu",&current[SWAP_IN],&current[SWAP_OUT]);
-       l=fgets(line,BUFFLEN,f); /* intr */
+       sscanf(line,"swap %llu %llu",&current[SWAP_IN],&current[SWAP_OUT]);
+       fgets_realloc(&line,&length,f); /* intr */
       }
 
     if(available[INTR])
       {
-       sscanf(line,"intr %lu",&current[INTR]);
-       l=fgets(line,BUFFLEN,f); /* disk_io or ctxt */
+       sscanf(line,"intr %llu",&current[INTR]);
+       fgets_realloc(&line,&length,f); /* disk_io or ctxt */
       }
 
     if(kernel_version_240 && available[DISK]) /* kernel version > ~2.4.0-test4 */
       {
        int num=8,nm,nr=0;
-       unsigned long d1,d3;
+       unsigned long long d1,d3;
 
        current[DISK_READ]=0;
        current[DISK_WRITE]=0;
@@ -489,9 +496,9 @@ int Update(time_t now,ProcMeterOutput *output)
           while(1)
             {
              if(kernel_version_240==6)
-                nr=sscanf(line+num," (%*d,%*d):(%*u,%lu,%*u,%lu)%n",&d1,&d3,&nm);
+                nr=sscanf(line+num," (%*d,%*d):(%*u,%llu,%*u,%llu)%n",&d1,&d3,&nm);
              else if(kernel_version_240==7)
-                nr=sscanf(line+num," (%*d,%*d):(%*u,%lu,%*u,%lu,%*u)%n",&d1,&d3,&nm);
+                nr=sscanf(line+num," (%*d,%*d):(%*u,%llu,%*u,%llu,%*u)%n",&d1,&d3,&nm);
 
              if(nr!=2)
                 break;
@@ -505,14 +512,18 @@ int Update(time_t now,ProcMeterOutput *output)
        if(available[DISK])
           current[DISK]=current[DISK_READ]+current[DISK_WRITE];
 
-       l=fgets(line,BUFFLEN,f); /* ctxt */
+       fgets_realloc(&line,&length,f); /* ctxt */
       }
 
     if(available[CONTEXT])
-       sscanf(line,"ctxt %lu",&current[CONTEXT]);
+       sscanf(line,"ctxt %llu",&current[CONTEXT]);
 
     if(available[CPU])
+      {
        current[CPU]=current[CPU_USER]+current[CPU_NICE]+current[CPU_SYS];
+       if(kernel_26)
+          current[CPU]+=cpu_iowait+cpu_irq+cpu_softirq+cpu_steal;
+      }
 
     if(available[PAGE])
        current[PAGE]=current[PAGE_IN]+current[PAGE_OUT];
@@ -576,4 +587,6 @@ int Update(time_t now,ProcMeterOutput *output)
 
 void Unload(void)
 {
+ if(line)
+    free(line);
 }
